#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from gbpservice.nfp.core import context as nfp_context
from gbpservice.nfp.core import controller as nfp_controller
from gbpservice.nfp.core import event as nfp_event
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import manager as nfp_manager
from gbpservice.nfp.core import worker as nfp_worker
import mock
import multiprocessing as multiprocessing
from oslo_config import cfg as oslo_config
from oslo_log import log as oslo_logging
import random
import six
import time
import unittest2
LOG = oslo_logging.getLogger(__name__)

NFP_MODULES_PATH = ['gbpservice.neutron.tests.unit.nfp.core']


def mocked_get_logging_context(**kwargs):
    return {
        'meta_id': '',
        'auth_token': None,
        'namespace': None}

nfp_logging.get_logging_context = mocked_get_logging_context


class MockedPipe(object):

    def __init__(self):
        self.fd = random.randint(14, 34)
        self.other_end_event_proc_func = None

    def poll(self, *args, **kwargs):
        return False

    def send(self, event):
        self.other_end_event_proc_func(event)


class MockedProcess(object):

    def __init__(self, parent_pipe=None, child_pipe=None,
                 controller=None):

        self.parent_pipe = parent_pipe
        self.child_pipe = child_pipe
        self.controller = controller
        self.daemon = True
        self.pid = random.randint(8888, 9999)

    def start(self):
        self.worker = nfp_worker.NfpWorker({}, threads=0)
        self.worker.parent_pipe = self.parent_pipe
        self.worker.pipe = self.child_pipe
        self.worker.controller = nfp_controller.NfpController(
            self.controller._conf, singleton=False)

        # fork a new controller object
        self.worker.controller.PROCESS_TYPE = "worker"
        self.worker.controller._pipe = self.worker.pipe
        self.worker.controller._event_handlers = (
            self.controller._event_handlers)
        self.worker.event_handlers = self.controller.get_event_handlers()

        self.parent_pipe.other_end_event_proc_func = (
            self.worker._process_event)
        self.child_pipe.other_end_event_proc_func = (
            self.controller._process_event)


def mocked_pipe(**kwargs):
    return MockedPipe(), MockedPipe()


def mocked_process(target=None, args=None):
    return MockedProcess(parent_pipe=args[1],
                         child_pipe=args[2],
                         controller=args[3])


nfp_controller.PIPE = mocked_pipe
nfp_controller.PROCESS = mocked_process


class MockedWatchdog(object):

    def __init__(self, handler, seconds=1, event=None):
        if event and event.desc.type == 'poll_event':
            # time.sleep(seconds)
            handler(event=event)

    def cancel(self):
        pass

nfp_manager.WATCHDOG = MockedWatchdog


class Object(object):

    def __init__(self):
        pass


class Test_Process_Model(unittest2.TestCase):

    def setUp(self):
        nfp_context.init()

    def _mocked_fork(self, args):
        proc = Object()
        pid = random.randint(8888, 9999)
        setattr(proc, 'pid', pid)
        return proc

    def _mocked_oslo_wrap(self):
        wrap = Object()
        setattr(wrap, 'service', {})
        return wrap

    def _mocked_event_ack(self, event):
        if event.id == 'TEST_EVENT_ACK_FROM_WORKER':
            if hasattr(event, 'desc'):
                if event.desc.worker:
                    self.controller.event_ack_wait_obj.set()

    def test_nfp_module_init(self):
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        wait_obj = multiprocessing.Event()
        setattr(controller, 'nfp_module_init_wait_obj', wait_obj)
        nfp_controller.load_nfp_modules(conf, controller)
        controller.nfp_module_init_wait_obj.wait(1)
        called = controller.nfp_module_init_wait_obj.is_set()
        self.assertTrue(called)

    def test_nfp_module_init_wrong_path(self):
        conf = oslo_config.CONF
        conf.nfp_modules_path = ['tmp.nfp']
        controller = nfp_controller.NfpController(oslo_config.CONF,
                                                  singleton=False)
        wait_obj = multiprocessing.Event()
        setattr(controller, 'nfp_module_init_wait_obj', wait_obj)
        nfp_controller.load_nfp_modules(conf, controller)
        controller.nfp_module_init_wait_obj.wait(1)
        called = controller.nfp_module_init_wait_obj.is_set()
        self.assertFalse(called)

    def test_nfp_module_post_init_called(self):
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        wait_obj = multiprocessing.Event()
        setattr(controller, 'nfp_module_post_init_wait_obj', wait_obj)
        nfp_modules = nfp_controller.load_nfp_modules(conf, controller)
        nfp_controller.nfp_modules_post_init(conf, nfp_modules, controller)
        controller.nfp_module_post_init_wait_obj.wait(1)
        called = controller.nfp_module_post_init_wait_obj.is_set()
        self.assertTrue(called)

    def test_nfp_module_post_init_ignored(self):
        # None the post_init method in test handler
        from gbpservice.neutron.tests.unit.nfp.core import (
            nfp_module)
        del nfp_module.nfp_module_post_init

        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        wait_obj = multiprocessing.Event()
        setattr(controller, 'nfp_module_post_init_wait_obj', wait_obj)
        nfp_modules = nfp_controller.load_nfp_modules(conf, controller)
        nfp_controller.nfp_modules_post_init(conf, nfp_modules, controller)
        controller.nfp_module_post_init_wait_obj.wait(1)
        called = controller.nfp_module_post_init_wait_obj.is_set()
        self.assertFalse(called)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController._fork'
    )
    def test_nfp_controller_launch_2_workers(self, mock_fork):
        mock_fork.side_effect = self._mocked_fork
        conf = oslo_config.CONF
        conf.nfp_modules_path = []
        controller = nfp_controller.NfpController(conf, singleton=False)
        controller.launch(2)
        # Check if 2 workers are created
        workers = controller.get_childrens()
        pids = workers.keys()
        self.assertTrue(len(pids) == 2)
        self.assertTrue(pid in range(8888, 9999) for pid in pids)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController._fork'
    )
    def test_nfp_controller_launch_4_workers(self, mock_fork):
        mock_fork.side_effect = self._mocked_fork
        conf = oslo_config.CONF
        conf.nfp_modules_path = []
        controller = nfp_controller.NfpController(conf, singleton=False)
        controller.launch(4)
        # Check if 4 workers are created
        workers = controller.get_childrens()
        pids = workers.keys()
        self.assertTrue(len(pids) == 4)
        self.assertTrue(pid in range(8888, 9999) for pid in pids)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController._fork'
    )
    def test_nfp_rsrc_manager_new_childs(self, mock_fork):
        mock_fork.side_effect = self._mocked_fork
        conf = oslo_config.CONF
        conf.nfp_modules_path = []
        controller = nfp_controller.NfpController(conf, singleton=False)
        controller.launch(2)
        controller._update_manager()
        # Check if 2 workers are added to manager
        pids = controller._manager._resource_map.keys()
        self.assertTrue(len(pids) == 2)
        self.assertTrue(pid in range(8888, 9999) for pid in pids)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController._fork'
    )
    def test_nfp_rsrc_manager_kill_child(self, mock_fork):
        mock_fork.side_effect = self._mocked_fork
        conf = oslo_config.CONF
        conf.nfp_modules_path = []
        controller = nfp_controller.NfpController(conf, singleton=False)
        controller.launch(2)
        controller._update_manager()
        # run so that it stores the snapshot
        controller._manager.manager_run()
        # Mock killing a child, remove it from workers list
        workers = controller.get_childrens()
        old_childs = list(workers.keys())
        del controller.children[old_childs[0]]
        # Mock creating a new child which replaces the killed one
        wrap = self._mocked_oslo_wrap()
        pid = controller.fork_child(wrap)
        controller.children[pid] = wrap

        # Run one more time and check if it detects the difference
        controller._manager.manager_run()
        pids = controller._manager._resource_map.keys()
        self.assertTrue(len(pids) == 2)
        if pid not in old_childs:
            self.assertFalse(old_childs[0] in pids)
        self.assertTrue(old_childs[1] in pids)

    def test_post_event_with_no_handler(self):
        conf = oslo_config.CONF
        conf.nfp_modules_path = []
        controller = nfp_controller.NfpController(conf, singleton=False)
        event = controller.create_event(
            id='EVENT_INVALID', data='INVALID_DATA',
            binding_key='EVENT_INVALID')
        try:
            controller.post_event(event)
        except AssertionError:
            return

        self.assertTrue(False)

    def mocked_compress(self, event):
        pass

    def mocked_pipe_send(self, pipe, event):
        if event.id == 'EVENT_1':
            if hasattr(event, 'desc'):
                if event.desc.worker:
                    self.controller.nfp_event_1_wait_obj.set()
        elif 'EVENT_LOAD' in event.id:
            if hasattr(event, 'desc'):
                if event.desc.worker == event.data:
                    self.controller.nfp_event_load_wait_obj.set()
        elif 'SEQUENCE' in event.id:
            if hasattr(event, 'desc'):
                if event.desc.worker:
                    if 'EVENT_1' in event.id:
                        self.controller.sequence_event_1_wait_obj.set()
                    elif 'EVENT_2' in event.id:
                        self.controller.sequence_event_2_wait_obj.set()
        elif 'POLL' in event.id:
            if hasattr(event, 'desc'):
                if hasattr(event.desc, 'poll_desc'):
                    if event.desc.worker:
                        if event.id == 'POLL_EVENT':
                            self.controller.poll_event_wait_obj.set()
                        if event.id == 'POLL_EVENT_DECORATOR':
                            self.controller.poll_event_dec_wait_obj.set()

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.pipe_send'
    )
    def test_post_event_in_distributor(self, mock_pipe_send):
        mock_pipe_send.side_effect = self.mocked_pipe_send
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        nfp_controller.load_nfp_modules(conf, controller)
        # Mock launching of a worker
        controller.launch(1)
        controller._update_manager()
        wait_obj = multiprocessing.Event()
        setattr(controller, 'nfp_event_1_wait_obj', wait_obj)
        event = controller.create_event(
            id='EVENT_1',
            data='post_event_in_distributor')

        # Store in class object
        self.controller = controller
        controller.post_event(event)
        controller.nfp_event_1_wait_obj.wait(1)
        called = controller.nfp_event_1_wait_obj.is_set()
        self.assertTrue(called)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.pipe_send')
    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.compress')
    def test_load_distribution_to_workers(self, mock_compress, mock_pipe_send):
        mock_pipe_send.side_effect = self.mocked_pipe_send
        mock_compress.side_effect = self.mocked_compress
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        self.controller = controller
        nfp_controller.load_nfp_modules(conf, controller)
        # Mock launching of a worker
        controller.launch(3)
        controller._update_manager()

        # Load distribution as -> worker1 - 2, worker2 - 4, worker3 - 6
        # 10 events to be distributed.
        # worker1 will get 5
        # worker2 will get 4
        # worker3 will get 1
        # At the end all workers should be @load 7

        # Initialize with above load
        init_load = [6, 4, 2]
        worker_pids = []
        resource_map = controller._manager._resource_map
        for pid, em in six.iteritems(resource_map):
            load = init_load.pop()
            em._load = load
            worker_pids.append(pid)

        events = [
            controller.create_event(id='EVENT_LOAD_1', data=worker_pids[0]),
            controller.create_event(id='EVENT_LOAD_2', data=worker_pids[0]),
            controller.create_event(id='EVENT_LOAD_3', data=worker_pids[0]),
            controller.create_event(id='EVENT_LOAD_4', data=worker_pids[1]),
            controller.create_event(id='EVENT_LOAD_5', data=worker_pids[0]),
            controller.create_event(id='EVENT_LOAD_6', data=worker_pids[1]),
            controller.create_event(id='EVENT_LOAD_7', data=worker_pids[0]),
            controller.create_event(id='EVENT_LOAD_8', data=worker_pids[1]),
            controller.create_event(id='EVENT_LOAD_9', data=worker_pids[2])]

        for i in range(0, 9):
            wait_obj = multiprocessing.Event()
            setattr(controller, 'nfp_event_load_wait_obj', wait_obj)
            event = events[i]
            controller.post_event(event)
            controller.nfp_event_load_wait_obj.wait(1)
            called = controller.nfp_event_load_wait_obj.is_set()
            self.assertTrue(called)

    def test_new_event_with_sequence_and_no_binding_key(self):
        conf = oslo_config.CONF
        conf.nfp_modules_path = []
        controller = nfp_controller.NfpController(conf, singleton=False)
        event = controller.create_event(
            id='EVENT_SEQUENCE', data='NO_DATA',
            serialize=True)
        self.assertTrue(event is None)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.pipe_send'
    )
    def test_events_sequencing_with_same_binding_key(self, mock_pipe_send):
        mock_pipe_send.side_effect = self.mocked_pipe_send
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        self.controller = controller
        nfp_controller.load_nfp_modules(conf, controller)
        # Mock launching of a worker
        controller.launch(1)
        controller._update_manager()
        self.controller = controller

        wait_obj = multiprocessing.Event()
        setattr(controller, 'sequence_event_1_wait_obj', wait_obj)
        wait_obj = multiprocessing.Event()
        setattr(controller, 'sequence_event_2_wait_obj', wait_obj)
        event_1 = controller.create_event(
            id='SEQUENCE_EVENT_1', data='NO_DATA',
            serialize=True, binding_key='SEQUENCE')
        event_2 = controller.create_event(
            id='SEQUENCE_EVENT_2', data='NO_DATA',
            serialize=True, binding_key='SEQUENCE')
        controller.post_event(event_1)
        controller.post_event(event_2)

        controller._manager.manager_run()
        controller.sequence_event_1_wait_obj.wait(1)
        called = controller.sequence_event_1_wait_obj.is_set()
        self.assertTrue(called)
        controller.event_complete(event_1)
        controller._manager.manager_run()
        controller.sequence_event_2_wait_obj.wait(1)
        called = controller.sequence_event_2_wait_obj.is_set()
        self.assertTrue(called)
        controller.event_complete(event_2)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.pipe_send'
    )
    def test_events_sequencing_with_diff_binding_key(self, mock_pipe_send):
        mock_pipe_send.side_effect = self.mocked_pipe_send
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        self.controller = controller
        nfp_controller.load_nfp_modules(conf, controller)
        # Mock launching of a worker
        controller.launch(1)
        controller._update_manager()
        self.controller = controller

        wait_obj = multiprocessing.Event()
        setattr(controller, 'sequence_event_1_wait_obj', wait_obj)
        wait_obj = multiprocessing.Event()
        setattr(controller, 'sequence_event_2_wait_obj', wait_obj)
        event_1 = controller.create_event(
            id='SEQUENCE_EVENT_1', data='NO_DATA',
            serialize=True, binding_key='SEQUENCE_1')
        event_2 = controller.create_event(
            id='SEQUENCE_EVENT_2', data='NO_DATA',
            serialize=True, binding_key='SEQUENCE_2')
        controller.post_event(event_1)
        controller.post_event(event_2)

        controller._manager.manager_run()
        controller.sequence_event_1_wait_obj.wait(1)
        called = controller.sequence_event_1_wait_obj.is_set()
        self.assertTrue(called)
        controller.sequence_event_2_wait_obj.wait(1)
        called = controller.sequence_event_2_wait_obj.is_set()
        self.assertTrue(called)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.pipe_send'
    )
    def test_events_sequencing_negative(self, mock_pipe_send):
        mock_pipe_send.side_effect = self.mocked_pipe_send
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        self.controller = controller
        nfp_controller.load_nfp_modules(conf, controller)
        # Mock launching of a worker
        controller.launch(1)
        controller._update_manager()
        self.controller = controller

        wait_obj = multiprocessing.Event()
        setattr(controller, 'sequence_event_1_wait_obj', wait_obj)
        wait_obj = multiprocessing.Event()
        setattr(controller, 'sequence_event_2_wait_obj', wait_obj)
        event_1 = controller.create_event(
            id='SEQUENCE_EVENT_1', data='NO_DATA',
            serialize=True, binding_key='SEQUENCE')
        event_2 = controller.create_event(
            id='SEQUENCE_EVENT_2', data='NO_DATA',
            serialize=True, binding_key='SEQUENCE')
        controller.post_event(event_1)
        controller.post_event(event_2)

        controller._manager.manager_run()
        controller.sequence_event_1_wait_obj.wait(1)
        called = controller.sequence_event_1_wait_obj.is_set()
        self.assertTrue(called)
        controller._manager.manager_run()
        controller.sequence_event_2_wait_obj.wait(1)
        called = controller.sequence_event_2_wait_obj.is_set()
        # Should not be called
        self.assertFalse(called)
        controller.event_complete(event_1)
        controller.event_complete(event_2)

        @mock.patch(
            'gbpservice.nfp.core.controller.NfpController.pipe_send')
        @mock.patch(
            'gbpservice.nfp.core.controller.NfpController.compress')
        def test_poll_event(self, mock_compress, mock_pipe_send):
            mock_pipe_send.side_effect = self.mocked_pipe_send
            mock_compress.side_effect = self.mocked_compress
            conf = oslo_config.CONF
            conf.nfp_modules_path = NFP_MODULES_PATH
            controller = nfp_controller.NfpController(conf, singleton=False)
            self.controller = controller
            nfp_controller.load_nfp_modules(conf, controller)
            # Mock launching of a worker
            controller.launch(1)
            controller._update_manager()
            self.controller = controller

        wait_obj = multiprocessing.Event()
        setattr(controller, 'poll_event_wait_obj', wait_obj)
        event = controller.create_event(
            id='POLL_EVENT', data='NO_DATA')

        # Update descriptor
        desc = nfp_event.EventDesc(**{})
        setattr(event, 'desc', desc)
        event.desc.worker = controller.get_childrens().keys()[0]

        ctx = nfp_context.get()
        ctx['log_context']['namespace'] = 'nfp_module'

        controller.poll_event(event, spacing=1)
        # controller._manager.manager_run()

        start_time = time.time()
        # relinquish for 1sec
        time.sleep(1)

        # controller.poll()
        controller.poll_event_wait_obj.wait(0.1)
        called = controller.poll_event_wait_obj.is_set()
        end_time = time.time()
        self.assertTrue(called)
        self.assertTrue(round(end_time - start_time) == 1.0)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.pipe_send')
    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.compress')
    def test_poll_event_with_no_worker(self, mock_compress, mock_pipe_send):
        mock_compress.side_effect = self.mocked_compress
        mock_pipe_send.side_effect = self.mocked_pipe_send
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        self.controller = controller
        nfp_controller.load_nfp_modules(conf, controller)
        # Mock launching of a worker
        controller.launch(1)
        controller._update_manager()
        self.controller = controller

        wait_obj = multiprocessing.Event()
        setattr(controller, 'poll_event_wait_obj', wait_obj)
        event = controller.create_event(
            id='POLL_EVENT', data='NO_DATA')

        # Update descriptor
        desc = nfp_event.EventDesc(**{})
        setattr(event, 'desc', desc)
        # Explicitly make it none
        event.desc.worker = None

        ctx = nfp_context.get()
        ctx['log_context']['namespace'] = 'nfp_module'

        controller.poll_event(event, spacing=1)
        # controller._manager.manager_run()

        start_time = time.time()
        # relinquish for 1sec
        time.sleep(1)

        # controller.poll()
        controller.poll_event_wait_obj.wait(0.1)
        called = controller.poll_event_wait_obj.is_set()
        end_time = time.time()
        self.assertTrue(called)
        self.assertTrue(round(end_time - start_time) == 1.0)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.pipe_send')
    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.compress')
    def test_poll_event_with_decorator_spacing(self,
                                               mock_compress, mock_pipe_send):

        mock_pipe_send.side_effect = self.mocked_pipe_send
        mock_compress.side_effect = self.mocked_compress
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        self.controller = controller
        nfp_controller.load_nfp_modules(conf, controller)
        # Mock launching of a worker
        controller.launch(1)
        controller._update_manager()
        self.controller = controller

        wait_obj = multiprocessing.Event()
        setattr(controller, 'poll_event_dec_wait_obj', wait_obj)
        event = controller.create_event(
            id='POLL_EVENT_DECORATOR', data='NO_DATA')

        # Update descriptor
        desc = nfp_event.EventDesc(**{})
        setattr(event, 'desc', desc)
        # Explicitly make it none
        event.desc.worker = None

        ctx = nfp_context.get()
        ctx['log_context']['namespace'] = 'nfp_module'
        controller.poll_event(event)
        # controller._manager.manager_run()

        start_time = time.time()
        # relinquish for 2secs
        time.sleep(2)

        # controller.poll()
        controller.poll_event_dec_wait_obj.wait(0.1)
        called = controller.poll_event_dec_wait_obj.is_set()
        end_time = time.time()
        self.assertTrue(called)
        self.assertTrue(round(end_time - start_time) == 2.0)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.compress')
    def test_poll_event_with_no_spacing(self, mock_compress):
        mock_compress.side_effect = self.mocked_compress
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        event = controller.create_event(
            id='POLL_EVENT_WITHOUT_SPACING', data='NO_DATA')

        # Update descriptor
        desc = nfp_event.EventDesc(**{})
        setattr(event, 'desc', desc)
        # Explicitly make it none
        event.desc.worker = None

        try:
            controller.poll_event(event)
        except AssertionError as aerr:
            if aerr.message == "No spacing specified for polling":
                self.assertTrue(True)
                return

        # self.assertTrue(False)
        self.assertTrue(True)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.compress')
    def test_poll_event_with_no_handler(self, mock_compress):
        mock_compress.side_effect = self.mocked_compress
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        event = controller.create_event(
            id='POLL_EVENT_WITHOUT_HANDLER', data='NO_DATA')

        # Update descriptor
        desc = nfp_event.EventDesc(**{})
        setattr(event, 'desc', desc)
        # Explicitly make it none
        event.desc.worker = None

        try:
            controller.poll_event(event, spacing=1)
        except AssertionError as aerr:
            if "No poll handler found for event" in aerr.message:
                self.assertTrue(True)
                return

        self.assertTrue(False)

    @mock.patch(
        'gbpservice.nfp.core.manager.NfpResourceManager._event_acked')
    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.compress')
    def test_event_ack_from_worker(self, mock_event_acked, mock_compress):
        mock_event_acked.side_effect = self._mocked_event_ack
        mock_compress.side_effect = self.mocked_compress
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        self.controller = controller
        nfp_controller.load_nfp_modules(conf, controller)
        # Mock launching of a worker
        controller.launch(1)
        controller._update_manager()
        self.controller = controller

        # Check if 1 worker is added to manager
        pids = controller._manager._resource_map.keys()
        self.assertTrue(len(pids) == 1)
        self.assertTrue(pid in range(8888, 9999) for pid in pids)

        wait_obj = multiprocessing.Event()
        setattr(controller, 'event_ack_wait_obj', wait_obj)
        wait_obj = multiprocessing.Event()
        setattr(controller, 'event_ack_handler_cb_obj', wait_obj)
        event = controller.create_event(
            id='TEST_EVENT_ACK_FROM_WORKER', data='NO_DATA')
        controller.post_event(event)
        controller._manager.manager_run()

        # wait for event to be acked
        controller.event_ack_wait_obj.wait(1)
        called = controller.event_ack_wait_obj.is_set()
        self.assertTrue(called)

        # Check if event handler callback is invoked
        controller.event_ack_handler_cb_obj.wait(1)
        called = controller.event_ack_handler_cb_obj.is_set()
        self.assertTrue(called)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.compress'
    )
    def test_post_event_from_worker(self, mock_compress):
        mock_compress.side_effect = self.mocked_compress
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        self.controller = controller
        nfp_controller.load_nfp_modules(conf, controller)
        # Mock launching of a worker
        controller.launch(1)
        controller._update_manager()
        self.controller = controller

        # Check if 1 worker is added to manager
        pids = controller._manager._resource_map.keys()
        self.assertTrue(len(pids) == 1)
        self.assertTrue(pid in range(8888, 9999) for pid in pids)

        wait_obj = multiprocessing.Event()
        setattr(controller, 'post_event_worker_wait_obj', wait_obj)
        event = controller.create_event(
            id='TEST_POST_EVENT_FROM_WORKER', data='NO_DATA')
        worker_process = controller._worker_process.values()[0]
        worker_process.worker.controller.post_event(event)

        controller._manager.manager_run()

        # Check if event handler callback is invoked
        controller.post_event_worker_wait_obj.wait(1)
        called = controller.post_event_worker_wait_obj.is_set()
        self.assertTrue(called)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.compress'
    )
    def test_poll_event_from_worker(self, mock_compress):
        mock_compress.side_effect = self.mocked_compress
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        self.controller = controller
        nfp_controller.load_nfp_modules(conf, controller)
        # Mock launching of a worker
        controller.launch(1)
        controller._update_manager()
        self.controller = controller

        # Check if 1 worker is added to manager
        pids = controller._manager._resource_map.keys()
        self.assertTrue(len(pids) == 1)
        self.assertTrue(pid in range(8888, 9999) for pid in pids)

        wait_obj = multiprocessing.Event()
        setattr(controller, 'poll_event_worker_wait_obj', wait_obj)
        wait_obj = multiprocessing.Event()
        setattr(controller, 'poll_event_poll_wait_obj', wait_obj)

        event = controller.create_event(
            id='TEST_POLL_EVENT_FROM_WORKER', data='NO_DATA')
        worker_process = controller._worker_process.values()[0]
        worker_process.worker.controller.post_event(event)

        controller._manager.manager_run()

        # Check if event handler callback is invoked
        controller.poll_event_worker_wait_obj.wait(1)
        called = controller.poll_event_worker_wait_obj.is_set()
        self.assertTrue(called)

        time.sleep(1)
        # controller.poll()

        controller.poll_event_poll_wait_obj.wait(1)
        called = controller.poll_event_poll_wait_obj.is_set()
        self.assertTrue(called)

    @mock.patch(
        'gbpservice.nfp.core.controller.NfpController.compress'
    )
    def test_poll_event_cancelled_from_worker(self, mock_compress):
        mock_compress.side_effect = self.mocked_compress
        conf = oslo_config.CONF
        conf.nfp_modules_path = NFP_MODULES_PATH
        controller = nfp_controller.NfpController(conf, singleton=False)
        self.controller = controller
        nfp_controller.load_nfp_modules(conf, controller)
        # Mock launching of a worker
        controller.launch(1)
        controller._update_manager()
        self.controller = controller

        # Check if 1 worker is added to manager
        pids = controller._manager._resource_map.keys()
        self.assertTrue(len(pids) == 1)
        self.assertTrue(pid in range(8888, 9999) for pid in pids)

        wait_obj = multiprocessing.Event()
        setattr(controller, 'poll_event_worker_wait_obj', wait_obj)
        wait_obj = multiprocessing.Event()
        setattr(controller, 'poll_event_poll_wait_obj', wait_obj)
        wait_obj = multiprocessing.Event()
        setattr(controller, 'poll_event_poll_cancel_wait_obj', wait_obj)

        event = controller.create_event(
            id='TEST_POLL_EVENT_CANCEL_FROM_WORKER', data='NO_DATA')
        worker_process = controller._worker_process.values()[0]
        worker_process.worker.controller.post_event(event)

        controller._manager.manager_run()

        # Check if event handler callback is invoked
        controller.poll_event_worker_wait_obj.wait(1)
        called = controller.poll_event_worker_wait_obj.is_set()
        self.assertTrue(called)

        time.sleep(1)
        # controller.poll()

        controller.poll_event_poll_wait_obj.wait(1)
        called = controller.poll_event_poll_wait_obj.is_set()
        self.assertTrue(called)

        time.sleep(1)
        # controller.poll()

        controller.poll_event_poll_wait_obj.wait(1)
        called = controller.poll_event_poll_wait_obj.is_set()
        self.assertTrue(called)

        controller.poll_event_poll_cancel_wait_obj.wait(1)
        called = controller.poll_event_poll_cancel_wait_obj.is_set()
        self.assertTrue(called)

if __name__ == '__main__':
    unittest2.main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log

from gbpservice.neutron.services.servicechain.plugins.ncp import driver_base


class NoopNodeDriver(driver_base.NodeDriverBase):

    initialized = False

    @log.log_method_call
    def initialize(self, name):
        self.initialized = True
        self._name = name

    @log.log_method_call
    def get_plumbing_info(self, context):
        pass

    @log.log_method_call
    def validate_create(self, context):
        pass

    @log.log_method_call
    def validate_update(self, context):
        pass

    @log.log_method_call
    def create(self, context):
        pass

    @log.log_method_call
    def delete(self, context):
        pass

    @log.log_method_call
    def update(self, context):
        pass

    @log.log_method_call
    def update_policy_target_added(self, context, policy_target):
        pass

    @log.log_method_call
    def update_policy_target_removed(self, context, policy_target):
        pass

    @log.log_method_call
    def update_node_consumer_ptg_added(self, context, policy_target_group):
        pass

    @log.log_method_call
    def update_node_consumer_ptg_removed(self, context, policy_target_group):
        pass

    @log.log_method_call
    def notify_chain_parameters_updated(self, context):
        pass

    @log.log_method_call
    def policy_target_group_updated(self, context, old_policy_target_group,
                                    current_policy_target_group):
        pass

    @property
    def name(self):
        return self._name

    @log.log_method_call
    def get_status(self, context):
        pass



# Copyright 2017 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""fix migrations

Revision ID: d978a7a73785
Revises: 27b724002081
Create Date: 2017-09-18 17:34:18.856803

"""

# revision identifiers, used by Alembic.
revision = 'd978a7a73785'
down_revision = '27b724002081'

import os
import sys

from neutron.db import migration

from oslo_utils import importutils

from gbpservice.neutron.db.migration import alembic_migrations as am

# This is a hack to get around the fact that the versions
# directory has no __init__.py
filepath = os.path.abspath(am.__file__)
basepath = filepath[:filepath.rfind("/")] + "/versions"
sys.path.append(basepath)

DB_5239b0a50036 = '5239b0a50036_rename_tenant_to_project'
DB_c460c5682e74 = 'c460c5682e74_nfp_rename_tenant_to_project'


def upgrade():
    ensure_5239b0a50036_migration()
    ensure_c460c5682e74_migration()
    # remove the appended path
    del sys.path[sys.path.index(basepath)]


def ensure_5239b0a50036_migration():
    if not migration.schema_has_column('gp_l2_policies',
                                       'project_id'):
        db_5239b0a50036 = importutils.import_module(DB_5239b0a50036)
        db_5239b0a50036.upgrade()


def ensure_c460c5682e74_migration():
    if not migration.schema_has_column('nfp_port_infos',
                                       'project_id'):
        db_c460c5682e74 = importutils.import_module(DB_c460c5682e74)
        db_c460c5682e74.upgrade()


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron.callbacks import events
from neutron.callbacks import registry
from neutron.callbacks import resources
from neutron.db import api as db_api
from neutron.db import common_db_mixin
from neutron.db import l3_db
from neutron.db.models import securitygroup as sg_models
from neutron.db import models_v2
from neutron.db import securitygroups_db
from neutron.extensions import securitygroup as ext_sg
from neutron.plugins.ml2 import db as ml2_db
from neutron_lib.api import validators
from neutron_lib import constants
from neutron_lib import exceptions as n_exc
from oslo_log import log
from oslo_utils import excutils
from oslo_utils import uuidutils
from sqlalchemy import event
from sqlalchemy.orm import exc


LOG = log.getLogger(__name__)


# REVISIT(ivar): Monkey patch to allow explicit router_id to be set in Neutron
# for Floating Ip creation (for internal calls only). Once we split the server,
# this could be part of a GBP Neutron L3 driver.
def _get_assoc_data(self, context, fip, floatingip_db):
    (internal_port, internal_subnet_id,
     internal_ip_address) = self._internal_fip_assoc_data(
         context, fip, floatingip_db['tenant_id'])
    if fip.get('router_id'):
        router_id = fip['router_id']
        del fip['router_id']
    else:
        router_id = self._get_router_for_floatingip(
            context, internal_port, internal_subnet_id,
            floatingip_db['floating_network_id'])

    return fip['port_id'], internal_ip_address, router_id


l3_db.L3_NAT_dbonly_mixin._get_assoc_data = _get_assoc_data


def _update_fip_assoc(self, context, fip, floatingip_db, external_port):
    previous_router_id = floatingip_db.router_id
    port_id, internal_ip_address, router_id = (
        self._check_and_get_fip_assoc(context, fip, floatingip_db))
    update = {'fixed_ip_address': internal_ip_address,
              'fixed_port_id': port_id,
              'router_id': router_id,
              'last_known_router_id': previous_router_id}
    if 'description' in fip:
        update['description'] = fip['description']
    floatingip_db.update(update)
    next_hop = None
    if router_id:
        router = self._get_router(context.elevated(), router_id)
        gw_port = router.gw_port
        if gw_port:
            for fixed_ip in gw_port.fixed_ips:
                addr = netaddr.IPAddress(fixed_ip.ip_address)
                if addr.version == constants.IP_VERSION_4:
                    next_hop = fixed_ip.ip_address
                    break
    args = {'fixed_ip_address': internal_ip_address,
            'fixed_port_id': port_id,
            'router_id': router_id,
            'last_known_router_id': previous_router_id,
            'floating_ip_address': floatingip_db.floating_ip_address,
            'floating_network_id': floatingip_db.floating_network_id,
            'next_hop': next_hop,
            'context': context}
    registry.notify(resources.FLOATING_IP,
                    events.AFTER_UPDATE,
                    self._update_fip_assoc,
                    **args)

l3_db.L3_NAT_dbonly_mixin._update_fip_assoc = _update_fip_assoc


# REVISIT(ivar): Neutron adds a tenant filter on SG lookup for a given port,
# this breaks our service chain plumbing model so for now we should monkey
# patch the specific method. A follow up with the Neutron team is needed to
# figure out the reason for this and how to proceed for future releases.
def _get_security_groups_on_port(self, context, port):
    """Check that all security groups on port belong to tenant.

    :returns: all security groups IDs on port belonging to tenant.
    """
    p = port['port']
    if not validators.is_attr_set(
            p.get(securitygroups_db.ext_sg.SECURITYGROUPS)):
        return
    if p.get('device_owner') and p['device_owner'].startswith('network:'):
        return

    port_sg = p.get(securitygroups_db.ext_sg.SECURITYGROUPS, [])
    filters = {'id': port_sg}
    valid_groups = set(g['id'] for g in
                       self.get_security_groups(context, fields=['id'],
                                                filters=filters))

    requested_groups = set(port_sg)
    port_sg_missing = requested_groups - valid_groups
    if port_sg_missing:
        raise securitygroups_db.ext_sg.SecurityGroupNotFound(
            id=', '.join(port_sg_missing))

    return requested_groups

securitygroups_db.SecurityGroupDbMixin._get_security_groups_on_port = (
    _get_security_groups_on_port)


# REVISIT(kent): Neutron doesn't pass the remote_group_id while creating the
# ingress rule for the default SG. It also doesn't pass the newly created SG
# for the PRECOMMIT_CREATE event. Note that we should remove this in Pike as
# upstream has fixed the bug there
def create_security_group(self, context, security_group, default_sg=False):
    """Create security group.

    If default_sg is true that means we are a default security group for
    a given tenant if it does not exist.
    """
    s = security_group['security_group']
    kwargs = {
        'context': context,
        'security_group': s,
        'is_default': default_sg,
    }

    self._registry_notify(resources.SECURITY_GROUP, events.BEFORE_CREATE,
                          exc_cls=ext_sg.SecurityGroupConflict, **kwargs)

    tenant_id = s['tenant_id']

    if not default_sg:
        self._ensure_default_security_group(context, tenant_id)

    with db_api.autonested_transaction(context.session):
        security_group_db = sg_models.SecurityGroup(id=s.get('id') or (
                                          uuidutils.generate_uuid()),
                                          description=s['description'],
                                          tenant_id=tenant_id,
                                          name=s['name'])
        context.session.add(security_group_db)
        if default_sg:
            context.session.add(sg_models.DefaultSecurityGroup(
                security_group=security_group_db,
                tenant_id=security_group_db['tenant_id']))
        for ethertype in ext_sg.sg_supported_ethertypes:
            if default_sg:
                # Allow intercommunication
                ingress_rule = sg_models.SecurityGroupRule(
                    id=uuidutils.generate_uuid(), tenant_id=tenant_id,
                    security_group=security_group_db,
                    direction='ingress',
                    ethertype=ethertype,
                    remote_group_id=security_group_db.id,
                    source_group=security_group_db)
                context.session.add(ingress_rule)

            egress_rule = sg_models.SecurityGroupRule(
                id=uuidutils.generate_uuid(), tenant_id=tenant_id,
                security_group=security_group_db,
                direction='egress',
                ethertype=ethertype)
            context.session.add(egress_rule)

        secgroup_dict = self._make_security_group_dict(security_group_db)
        kwargs['security_group'] = secgroup_dict
        self._registry_notify(resources.SECURITY_GROUP,
                              events.PRECOMMIT_CREATE,
                              exc_cls=ext_sg.SecurityGroupConflict,
                              **kwargs)

    registry.notify(resources.SECURITY_GROUP, events.AFTER_CREATE, self,
                    **kwargs)
    return secgroup_dict

securitygroups_db.SecurityGroupDbMixin.create_security_group = (
    create_security_group)


# REVISIT(kent): Neutron doesn't pass the updated SG for the PRECOMMIT_UPDATE
# event. Note that we should remove this in Pike as upstream has fixed the bug
# there
def update_security_group(self, context, id, security_group):
    s = security_group['security_group']

    kwargs = {
        'context': context,
        'security_group_id': id,
        'security_group': s,
    }
    self._registry_notify(resources.SECURITY_GROUP, events.BEFORE_UPDATE,
                          exc_cls=ext_sg.SecurityGroupConflict, **kwargs)

    with context.session.begin(subtransactions=True):
        sg = self._get_security_group(context, id)
        if sg['name'] == 'default' and 'name' in s:
            raise ext_sg.SecurityGroupCannotUpdateDefault()
        sg_dict = self._make_security_group_dict(sg)
        kwargs['original_security_group'] = sg_dict
        sg.update(s)
        sg_dict = self._make_security_group_dict(sg)
        kwargs['security_group'] = sg_dict
        self._registry_notify(
                resources.SECURITY_GROUP,
                events.PRECOMMIT_UPDATE,
                exc_cls=ext_sg.SecurityGroupConflict, **kwargs)

    registry.notify(resources.SECURITY_GROUP, events.AFTER_UPDATE, self,
                    **kwargs)
    return sg_dict

securitygroups_db.SecurityGroupDbMixin.update_security_group = (
    update_security_group)


# REVISIT(kent): Neutron doesn't pass the SG rules for the PRECOMMIT_DELETE
# event. Note that we should remove this in Pike as upstream has fixed the bug
# there
def delete_security_group(self, context, id):
    filters = {'security_group_id': [id]}
    ports = self._get_port_security_group_bindings(context, filters)
    if ports:
        raise ext_sg.SecurityGroupInUse(id=id)
    # confirm security group exists
    sg = self._get_security_group(context, id)

    if sg['name'] == 'default' and not context.is_admin:
        raise ext_sg.SecurityGroupCannotRemoveDefault()
    kwargs = {
        'context': context,
        'security_group_id': id,
        'security_group': sg,
    }
    self._registry_notify(resources.SECURITY_GROUP, events.BEFORE_DELETE,
                          exc_cls=ext_sg.SecurityGroupInUse, id=id,
                          **kwargs)

    with context.session.begin(subtransactions=True):
        # pass security_group_rule_ids to ensure
        # consistency with deleted rules
        kwargs['security_group_rule_ids'] = [r['id'] for r in sg.rules]
        kwargs['security_group'] = self._make_security_group_dict(sg)
        self._registry_notify(resources.SECURITY_GROUP,
                              events.PRECOMMIT_DELETE,
                              exc_cls=ext_sg.SecurityGroupInUse, id=id,
                              **kwargs)
        context.session.delete(sg)

    kwargs.pop('security_group')
    registry.notify(resources.SECURITY_GROUP, events.AFTER_DELETE, self,
                    **kwargs)

securitygroups_db.SecurityGroupDbMixin.delete_security_group = (
    delete_security_group)


# REVISIT(kent): Neutron doesn't pass the newly created SG rule for the
# PRECOMMIT_CREATE event. Note that we should remove this in Pike as upstream
# has fixed the bug there
def _create_security_group_rule(self, context, security_group_rule,
                                validate=True):
    if validate:
        self._validate_security_group_rule(context, security_group_rule)
    rule_dict = security_group_rule['security_group_rule']
    kwargs = {
        'context': context,
        'security_group_rule': rule_dict
    }
    self._registry_notify(resources.SECURITY_GROUP_RULE,
                          events.BEFORE_CREATE,
                          exc_cls=ext_sg.SecurityGroupConflict, **kwargs)

    with context.session.begin(subtransactions=True):
        if validate:
            self._check_for_duplicate_rules_in_db(context,
                                                  security_group_rule)
        db = sg_models.SecurityGroupRule(
            id=(rule_dict.get('id') or uuidutils.generate_uuid()),
            tenant_id=rule_dict['tenant_id'],
            security_group_id=rule_dict['security_group_id'],
            direction=rule_dict['direction'],
            remote_group_id=rule_dict.get('remote_group_id'),
            ethertype=rule_dict['ethertype'],
            protocol=rule_dict['protocol'],
            port_range_min=rule_dict['port_range_min'],
            port_range_max=rule_dict['port_range_max'],
            remote_ip_prefix=rule_dict.get('remote_ip_prefix'),
            description=rule_dict.get('description')
        )
        context.session.add(db)
        res_rule_dict = self._make_security_group_rule_dict(db)
        kwargs['security_group_rule'] = res_rule_dict
        self._registry_notify(resources.SECURITY_GROUP_RULE,
                          events.PRECOMMIT_CREATE,
                          exc_cls=ext_sg.SecurityGroupConflict, **kwargs)
    registry.notify(
        resources.SECURITY_GROUP_RULE, events.AFTER_CREATE, self,
        **kwargs)
    return res_rule_dict

securitygroups_db.SecurityGroupDbMixin._create_security_group_rule = (
    _create_security_group_rule)


# REVISIT(kent): Neutron doesn't pass the SG ID of the rule for the
# PRECOMMIT_DELETE event. Note that we should remove this in Pike as upstream
# has fixed the bug there
def delete_security_group_rule(self, context, id):
    kwargs = {
        'context': context,
        'security_group_rule_id': id
    }
    self._registry_notify(resources.SECURITY_GROUP_RULE,
                          events.BEFORE_DELETE, id=id,
                          exc_cls=ext_sg.SecurityGroupRuleInUse, **kwargs)

    with context.session.begin(subtransactions=True):
        query = self._model_query(context,
                                  sg_models.SecurityGroupRule).filter(
            sg_models.SecurityGroupRule.id == id)
        try:
            # As there is a filter on a primary key it is not possible for
            # MultipleResultsFound to be raised
            sg_rule = query.one()
        except exc.NoResultFound:
            raise ext_sg.SecurityGroupRuleNotFound(id=id)

        kwargs['security_group_id'] = sg_rule['security_group_id']
        self._registry_notify(resources.SECURITY_GROUP_RULE,
                              events.PRECOMMIT_DELETE,
                              exc_cls=ext_sg.SecurityGroupRuleInUse, id=id,
                              **kwargs)
        context.session.delete(sg_rule)

    registry.notify(
        resources.SECURITY_GROUP_RULE, events.AFTER_DELETE, self,
        **kwargs)

securitygroups_db.SecurityGroupDbMixin.delete_security_group_rule = (
    delete_security_group_rule)


def get_port_from_device_mac(context, device_mac):
    LOG.debug("get_port_from_device_mac() called for mac %s", device_mac)
    qry = context.session.query(models_v2.Port).filter_by(
        mac_address=device_mac).order_by(models_v2.Port.device_owner.desc())
    return qry.first()

ml2_db.get_port_from_device_mac = get_port_from_device_mac

PUSH_NOTIFICATIONS_METHOD = None
DISCARD_NOTIFICATIONS_METHOD = None


def get_session(autocommit=True, expire_on_commit=False, use_slave=False):
    # The folowing are declared as global so that they can
    # used in the inner functions that follow.
    global PUSH_NOTIFICATIONS_METHOD
    global DISCARD_NOTIFICATIONS_METHOD
    from gbpservice.network.neutronv2 import local_api
    PUSH_NOTIFICATIONS_METHOD = (
        local_api.post_notifications_from_queue)
    DISCARD_NOTIFICATIONS_METHOD = (
        local_api.discard_notifications_after_rollback)

    # The following two lines are copied from the original
    # implementation of db_api.get_session() and should be updated
    # if the original implementation changes.
    new_session = db_api.context_manager.get_legacy_facade().get_session(
        autocommit=autocommit, expire_on_commit=expire_on_commit,
        use_slave=use_slave)

    new_session.notification_queue = {}

    def gbp_after_transaction(session, transaction):
        if transaction and not transaction._parent and (
            not transaction.is_active and not transaction.nested):
            if transaction in session.notification_queue:
                # push the queued notifications only when the
                # outermost transaction completes
                PUSH_NOTIFICATIONS_METHOD(session, transaction)

    def gbp_after_rollback(session):
        # We discard all queued notifiactions if the transaction fails.
        DISCARD_NOTIFICATIONS_METHOD(session)

    if local_api.QUEUE_OUT_OF_PROCESS_NOTIFICATIONS:
        event.listen(new_session, "after_transaction_end",
                     gbp_after_transaction)
        event.listen(new_session, "after_rollback",
                     gbp_after_rollback)

    return new_session


db_api.get_session = get_session


# REVISIT: This is temporary, the correct fix is to use
# the 'project_id' directly from the context rather than
# calling this method.
def _get_tenant_id_for_create(self, context, resource):
    if context.is_admin and 'tenant_id' in resource:
        tenant_id = resource['tenant_id']
    elif ('tenant_id' in resource and
          resource['tenant_id'] != context.project_id):
        reason = _('Cannot create resource for another tenant')
        raise n_exc.AdminRequired(reason=reason)
    else:
        tenant_id = context.project_id

    return tenant_id


common_db_mixin.CommonDbMixin._get_tenant_id_for_create = (
    _get_tenant_id_for_create)

# TODO(ivar): while this block would be better place in the patch_neutron
# module, it seems like being part of an "extension" package is the only
# way to make it work at the moment. Tests have shown that Neutorn reloads
# the extensions at every call (at least in the UTs) and this causes the
# AIM_FLC_L7_PARAMS to be reset over and over. By patching at this point,
# we make sure we always have the proper value for that variable.
try:
    import six
    import sys

    from networking_sfc.db import flowclassifier_db
    from networking_sfc.db import sfc_db
    from networking_sfc.extensions import flowclassifier as fc_ext
    from networking_sfc.extensions import sfc as sfc_ext  # noqa
    from networking_sfc.services.flowclassifier.common import (
        exceptions as fc_exc)
    from networking_sfc.services.flowclassifier import driver_manager as fc_mgr
    from networking_sfc.services.flowclassifier import plugin as fc_plugin
    from networking_sfc.services.sfc.common import exceptions as sfc_exc
    from networking_sfc.services.sfc import driver_manager as sfc_mgr
    from networking_sfc.services.sfc import plugin as sfc_plugin
    from neutron.services.trunk import constants as tcst
    from oslo_utils import uuidutils

    from gbpservice.neutron.services.sfc.aim import constants as sfc_cts

    if 'flowclassifier' in sys.modules:
        sys.modules['flowclassifier'].SUPPORTED_L7_PARAMETERS.update(
            sfc_cts.AIM_FLC_L7_PARAMS)
    if 'networking_sfc.extensions.flowclassifier' in sys.modules:
        sys.modules[
            ('networking_sfc.extensions.'
             'flowclassifier')].SUPPORTED_L7_PARAMETERS.update(
            sfc_cts.AIM_FLC_L7_PARAMS)
    if 'sfc' in sys.modules:
        sys.modules['sfc'].RESOURCE_ATTRIBUTE_MAP['port_pair_groups'][
            'port_pair_group_parameters']['validate']['type:dict'].update(
                sfc_cts.AIM_PPG_PARAMS)
    if 'networking_sfc.extensions.sfc' in sys.modules:
        sys.modules['networking_sfc.extensions.sfc'].RESOURCE_ATTRIBUTE_MAP[
            'port_pair_groups']['port_pair_group_parameters']['validate'][
            'type:dict'].update(sfc_cts.AIM_PPG_PARAMS)
    # REVISIT(ivar): The following diff will fix flow classifier creation
    # method when using L7 parameters.
    # -            key: L7Parameter(key, val)
    # +            key: L7Parameter(keyword=key, value=val)

    def create_flow_classifier(self, context, flow_classifier):
        fc = flow_classifier['flow_classifier']
        tenant_id = fc['tenant_id']
        l7_parameters = {
            key: flowclassifier_db.L7Parameter(keyword=key, value=val)
            for key, val in six.iteritems(fc['l7_parameters'])}
        ethertype = fc['ethertype']
        protocol = fc['protocol']
        source_port_range_min = fc['source_port_range_min']
        source_port_range_max = fc['source_port_range_max']
        self._check_port_range_valid(source_port_range_min,
                                     source_port_range_max,
                                     protocol)
        destination_port_range_min = fc['destination_port_range_min']
        destination_port_range_max = fc['destination_port_range_max']
        self._check_port_range_valid(destination_port_range_min,
                                     destination_port_range_max,
                                     protocol)
        source_ip_prefix = fc['source_ip_prefix']
        self._check_ip_prefix_valid(source_ip_prefix, ethertype)
        destination_ip_prefix = fc['destination_ip_prefix']
        self._check_ip_prefix_valid(destination_ip_prefix, ethertype)
        logical_source_port = fc['logical_source_port']
        logical_destination_port = fc['logical_destination_port']
        with context.session.begin(subtransactions=True):
            if logical_source_port is not None:
                self._get_port(context, logical_source_port)
            if logical_destination_port is not None:
                self._get_port(context, logical_destination_port)
            query = self._model_query(
                context, flowclassifier_db.FlowClassifier)
            for flow_classifier_db in query.all():
                if self.flowclassifier_conflict(
                    fc,
                    flow_classifier_db
                ):
                    raise fc_ext.FlowClassifierInConflict(
                        id=flow_classifier_db['id'])
            flow_classifier_db = flowclassifier_db.FlowClassifier(
                id=uuidutils.generate_uuid(),
                tenant_id=tenant_id,
                name=fc['name'],
                description=fc['description'],
                ethertype=ethertype,
                protocol=protocol,
                source_port_range_min=source_port_range_min,
                source_port_range_max=source_port_range_max,
                destination_port_range_min=destination_port_range_min,
                destination_port_range_max=destination_port_range_max,
                source_ip_prefix=source_ip_prefix,
                destination_ip_prefix=destination_ip_prefix,
                logical_source_port=logical_source_port,
                logical_destination_port=logical_destination_port,
                l7_parameters=l7_parameters
            )
            context.session.add(flow_classifier_db)
            return self._make_flow_classifier_dict(flow_classifier_db)
    flowclassifier_db.FlowClassifierDbPlugin.create_flow_classifier = (
        create_flow_classifier)

    # Flowclassifier validation should also take into account l7_parameters.

    old_validation = (
        flowclassifier_db.FlowClassifierDbPlugin.flowclassifier_basic_conflict)

    def flowclassifier_basic_conflict(cls, first_flowclassifier,
                                      second_flowclassifier):
        def _l7_params_conflict(fc1, fc2):
            if (validators.is_attr_set(fc1['l7_parameters']) and
                    validators.is_attr_set(fc2['l7_parameters'])):
                if fc1['l7_parameters'] == fc2['l7_parameters']:
                    return True
            return all(not validators.is_attr_set(fc['l7_parameters'])
                       for fc in [fc1, fc2])
        return cls._old_flowclassifier_basic_conflict(
            first_flowclassifier, second_flowclassifier) and (
            _l7_params_conflict(first_flowclassifier, second_flowclassifier))

    if getattr(flowclassifier_db.FlowClassifierDbPlugin,
               '_old_flowclassifier_basic_conflict', None) is None:
        flowclassifier_db.FlowClassifierDbPlugin.\
            _old_flowclassifier_basic_conflict = old_validation
        flowclassifier_db.FlowClassifierDbPlugin.\
            flowclassifier_basic_conflict = classmethod(
                flowclassifier_basic_conflict)

    # NOTE(ivar): Trunk subports don't have a device ID, we need this
    # validation to pass
    # NOTE(ivar): It would be ideal to re-use the original function and call
    # it instead of copying it here. However, it looks like this module gets
    # reloaded a number of times and this would cause the original definition
    # to be overridden with the new one, thus causing an endless recursion.
    def _validate_port_pair_ingress_egress(self, ingress, egress):
        if any(port.get('device_owner') == tcst.TRUNK_SUBPORT_OWNER
               for port in [ingress, egress]):
            return
        if 'device_id' not in ingress or not ingress['device_id']:
            raise sfc_db.ext_sfc.PortPairIngressNoHost(
                ingress=ingress['id']
            )
        if 'device_id' not in egress or not egress['device_id']:
            raise sfc_db.ext_sfc.PortPairEgressNoHost(
                egress=egress['id']
            )
        if ingress['device_id'] != egress['device_id']:
            raise sfc_db.ext_sfc.PortPairIngressEgressDifferentHost(
                ingress=ingress['id'],
                egress=egress['id'])
    sfc_db.SfcDbPlugin._validate_port_pair_ingress_egress = (
        _validate_port_pair_ingress_egress)

    if not getattr(sfc_plugin.SfcPlugin, '_patch_db_retry', False):
        sfc_plugin.SfcPlugin.create_port_pair = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.create_port_pair))
        sfc_plugin.SfcPlugin.create_port_pair_group = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.create_port_pair_group))
        sfc_plugin.SfcPlugin.create_port_chain = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.create_port_chain))

        sfc_plugin.SfcPlugin.update_port_pair = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.update_port_pair))
        sfc_plugin.SfcPlugin.update_port_pair_group = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.update_port_pair_group))
        sfc_plugin.SfcPlugin.update_port_chain = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.update_port_chain))

        sfc_plugin.SfcPlugin.delete_port_pair = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.delete_port_pair))
        sfc_plugin.SfcPlugin.delete_port_pair_group = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.delete_port_pair_group))
        sfc_plugin.SfcPlugin.delete_port_chain = (
            db_api.retry_if_session_inactive()(
                sfc_plugin.SfcPlugin.delete_port_chain))
        sfc_plugin.SfcPlugin._patch_db_retry = True

    if not getattr(fc_plugin.FlowClassifierPlugin, '_patch_db_retry', False):
        fc_plugin.FlowClassifierPlugin.create_flow_classifier = (
            db_api.retry_if_session_inactive()(
                fc_plugin.FlowClassifierPlugin.create_flow_classifier))
        fc_plugin.FlowClassifierPlugin.update_flow_classifier = (
            db_api.retry_if_session_inactive()(
                fc_plugin.FlowClassifierPlugin.update_flow_classifier))
        fc_plugin.FlowClassifierPlugin.delete_flow_classifier = (
            db_api.retry_if_session_inactive()(
                fc_plugin.FlowClassifierPlugin.delete_flow_classifier))
        fc_plugin.FlowClassifierPlugin._patch_db_retry = True

    # Overrides SFC implementation to avoid eating retriable exceptions
    def get_call_drivers(exception, name):
        def _call_drivers(self, method_name, context, raise_orig_exc=False):
            for driver in self.ordered_drivers:
                try:
                    getattr(driver.obj, method_name)(context)
                except Exception as e:
                    # This is an internal failure.
                    if db_api.is_retriable(e):
                        with excutils.save_and_reraise_exception():
                            LOG.debug(
                                "DB exception raised by extension driver "
                                "'%(name)s' in %(method)s",
                                {'name': driver.name, 'method': method_name},
                                exc_info=e)
                    LOG.exception(e)
                    LOG.error("%(plugin)s driver '%(name)s' "
                              "failed in %(method)s",
                              {'name': driver.name, 'method': method_name,
                               'plugin': name})
                    if raise_orig_exc:
                        raise
                    else:
                        raise exception(method=method_name)
        return _call_drivers

    sfc_mgr.SfcDriverManager._call_drivers = get_call_drivers(
        sfc_exc.SfcDriverError, 'SFC')
    fc_mgr.FlowClassifierDriverManager._call_drivers = get_call_drivers(
        fc_exc.FlowClassifierDriverError, 'FlowClassifier')

except ImportError as e:
    LOG.warning("Import error while patching networking-sfc: %s",
                e.message)



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_inject_default_route_attr_to_l2p

Revision ID: 7ef98f287d6
Create Date: 2015-11-29 02:00:06.217136

"""

# revision identifiers, used by Alembic.
revision = '7ef98f287d6'
down_revision = '478d86c6c648'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('gp_l2_policies',
                  sa.Column('inject_default_route', sa.Boolean,
                            server_default=sa.sql.true()))


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pecan

from gbpservice.contrib.nfp.configurator.advanced_controller import (
    controller)

"""This class forwards HTTP request to controller class.

This class create an object of Controller class with appropriate
parameter according to the path of HTTP request. According to the
parameter passed to Controller class it sends an RPC call/cast to
configurator.

"""


class ControllerResolver(object):

    create_network_function_device_config = controller.Controller(
        "create_network_function_device_config")
    delete_network_function_device_config = controller.Controller(
        "delete_network_function_device_config")
    update_network_function_device_config = controller.Controller(
        "update_network_function_device_config")
    create_network_function_config = controller.Controller(
        "create_network_function_config")
    delete_network_function_config = controller.Controller(
        "delete_network_function_config")
    update_network_function_config = controller.Controller(
        "update_network_function_config")
    get_notifications = controller.Controller("get_notifications")
    network_function_event = controller.Controller("network_function_event")
    get_requests = controller.Controller("get_requests")

""" This class forwards HTTP requests starting with /v1/nfp.

All HTTP requests with path starting from /v1
land here. This class forward request with path starting from /v1/nfp
to ControllerResolver.

"""


class V1Controller(object):

    nfp = ControllerResolver()

    @pecan.expose()
    def get(self):
        return {'versions': [{'status': 'CURRENT',
                              'updated': '2014-12-11T00:00:00Z',
                              'id': 'v1'}]}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcase:
    test = test_gbp_ri_func_4()
    test.run()


class test_gbp_ri_func_4(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_ri_func_4.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_ri_func_4.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.spec_name = 'demo_sc_spec'
        self.fw_name = 'demo_fw'
        self.lb_name = 'demo_lb'

    def cleanup(self, fail=0):
        for obj in ['node', 'spec']:
            self.gbpcfg.gbp_del_all_anyobj(obj)
        if fail != 0:
            self._log.info("\n## TESTCASE_GBP_RI_FUNC_4: FAILED")
            utils_libs.report_results('test_gbp_ri_func_4', 'test_results.txt')
            sys.exit(1)

    def run(self):
        self._log.info(
            "\n## TESTCASE_GBP_RI_FUNC_4: RESOURCE INTEGRITY AMONG "
            "SERVICE-CHAIN OBJECTS")
        # Testcase work-flow starts
        # ============ ALL POLICY OBJECTS ARE TO BE CREATED AND VERIFIED =
        self._log.info("\n##  Step 1: Create Service Chain Nodes LB & FW ##\n")
        lb_uuid = self.gbpcfg.gbp_sc_cfg_all(1, 'node', self.lb_name)
        if lb_uuid == 0:
            self._log.info(
                "# Step 1: Create Service Chain Loadbalance Node == Failed")
            self.cleanup(fail=1)
        fw_uuid = self.gbpcfg.gbp_sc_cfg_all(1, 'node', self.lb_name)
        if fw_uuid == 0:
            self._log.info(
                "# Step 1A: Create Service Chain Firewall Node == Failed")
            self.cleanup(fail=1)
        ######
        self._log.info("\n## Step 2: Create ServiceChain Spec ##\n")
        # Ensur that node names or node uuids passed as val to param
        # 'nodes',MUST be in order of FW and then LB.. this order is required
        # from gbp pov
        spec_uuid = self.gbpcfg.gbp_sc_cfg_all(
            1, 'spec', self.spec_name, nodes='%s %s' %
            (fw_uuid, lb_uuid))
        if spec_uuid == 0:
            self._log.info("# Step 2: Create ServiceChain Spec == Failed")
            self.cleanup(fail=1)
        ######
        self._log.info("\n## Step 3: Delete the Service Chain Nodes ##\n")
        cnt = 0
        for nodeid in [lb_uuid, fw_uuid]:
            if self.gbpcfg.gbp_sc_cfg_all(0, 'node', nodeid) != 0:
                self._log.info(
                    "# Step 4: Deletion of ServiceChain did NOT fail")
                cnt += 1
        if cnt > 0:
            self.cleanup(fail=1)
        else:
            self._log.info("\n## TESTCASE_GBP_RI_FUNC_4: PASSED")
            self.cleanup()
        utils_libs.report_results('test_gbp_ri_func_4', 'test_results.txt')
        sys.exit(1)

if __name__ == '__main__':
    main()



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# The following is imported at the beginning to ensure
# that the patches are applied before any of the
# modules save a reference to the functions being patched
from gbpservice._i18n import _LE
from gbpservice._i18n import _LI
from gbpservice.neutron.extensions import patch  # noqa
from gbpservice.neutron.plugins.ml2plus import patch_neutron  # noqa

import functools

from neutron.api.v2 import attributes
from neutron.callbacks import events
from neutron.callbacks import registry
from neutron.callbacks import resources
from neutron.common import constants as n_const
from neutron.db import api as db_api
from neutron.db import db_base_plugin_v2
from neutron.db.models import securitygroup as securitygroups_db
from neutron.db import models_v2
from neutron.db import provisioning_blocks
from neutron.extensions import address_scope as as_ext
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2 import managers as ml2_managers
from neutron.plugins.ml2 import plugin as ml2_plugin
from neutron.quota import resource_registry
from neutron_lib.api import validators
from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils

from gbpservice.neutron.db import implicitsubnetpool_db
from gbpservice.neutron.plugins.ml2plus import driver_api as api_plus
from gbpservice.neutron.plugins.ml2plus import driver_context
from gbpservice.neutron.plugins.ml2plus import managers

LOG = log.getLogger(__name__)


opts = [
    cfg.BoolOpt('refresh_network_db_obj',
                default=False,
                help=_("Refresh the network DB object to correctly "
                       "reflect the most recent state of all its "
                       "attributes. This refresh will be performed "
                       "in the _ml2_md_extend_network_dict method "
                       "inside the ml2plus plugin. The refresh option "
                       "may have a significant performace impact "
                       "and should be avoided. Hence this configuration "
                       "is set to False by default.")),
    cfg.BoolOpt('refresh_port_db_obj',
                default=False,
                help=_("Refresh the port DB object to correctly "
                       "reflect the most recent state of all its "
                       "attributes. This refresh will be performed "
                       "in the _ml2_md_extend_port_dict method "
                       "inside the ml2plus plugin. The refresh option "
                       "may have a significant performace impact "
                       "and should be avoided. Hence this configuration "
                       "is set to False by default.")),
    cfg.BoolOpt('refresh_subnet_db_obj',
                default=False,
                help=_("Refresh the subnet DB object to correctly "
                       "reflect the most recent state of all its "
                       "attributes. This refresh will be performed "
                       "in the _ml2_md_extend_subnet_dict method "
                       "inside the ml2plus plugin. The refresh option "
                       "may have a significant performace impact "
                       "and should be avoided. Hence this configuration "
                       "is set to False by default.")),
    cfg.BoolOpt('refresh_subnetpool_db_obj',
                default=False,
                help=_("Refresh the subnetpool DB object to correctly "
                       "reflect the most recent state of all its "
                       "attributes. This refresh will be performed "
                       "in the _ml2_md_extend_subnetpool_dict method "
                       "inside the ml2plus plugin. The refresh option "
                       "may have a significant performace impact "
                       "and should be avoided. Hence this configuration "
                       "is set to False by default.")),
    cfg.BoolOpt('refresh_address_scope_db_obj',
                default=False,
                help=_("Refresh the address_scope DB object to correctly "
                       "reflect the most recent state of all its "
                       "attributes. This refresh will be performed "
                       "in the _ml2_md_extend_address_scope_dict method "
                       "inside the ml2plus plugin. The refresh option "
                       "may have a significant performace impact "
                       "and should be avoided. Hence this configuration "
                       "is set to False by default.")),
]

cfg.CONF.register_opts(opts, "ml2plus")


def disable_transaction_guard(f):
    # We do not want to enforce transaction guard
    @functools.wraps(f)
    def inner(self, context, *args, **kwargs):
        setattr(context, 'GUARD_TRANSACTION', False)
        return f(self, context, *args, **kwargs)
    return inner


class Ml2PlusPlugin(ml2_plugin.Ml2Plugin,
                    implicitsubnetpool_db.ImplicitSubnetpoolMixin):

    """Extend the ML2 core plugin with missing functionality.

    The standard ML2 core plugin in Neutron is missing a few features
    needed for optimal APIC AIM support. This class adds those
    features, while maintaining compatibility with all standard ML2
    drivers and configuration. The only change necessary to use
    ML2Plus is to register the ml2plus entry point instead of the ml2
    entry port as Neutron's core_plugin. Drivers that need these
    features inherit from the extended MechanismDriver and
    ExtensionDriver abstract base classes.
    """

    __native_bulk_support = True
    __native_pagination_support = True
    __native_sorting_support = True
    ml2_plugin.Ml2Plugin._supported_extension_aliases += [
        "implicit-subnetpools"]

    # Override and bypass immediate base class's __init__ in order to
    # instantate extended manager class(es).
    @resource_registry.tracked_resources(
        network=models_v2.Network,
        port=models_v2.Port,
        subnet=models_v2.Subnet,
        subnetpool=models_v2.SubnetPool,
        security_group=securitygroups_db.SecurityGroup,
        security_group_rule=securitygroups_db.SecurityGroupRule)
    def __init__(self):
        LOG.info(_LI("Ml2Plus initializing"))
        registry._get_callback_manager()._notify_loop = (
            patch_neutron._notify_loop)
        # First load drivers, then initialize DB, then initialize drivers
        self.type_manager = ml2_managers.TypeManager()
        self.extension_manager = managers.ExtensionManager()
        self.mechanism_manager = managers.MechanismManager()
        super(ml2_plugin.Ml2Plugin, self).__init__()
        self.type_manager.initialize()
        self.extension_manager.initialize()
        self.mechanism_manager.initialize()
        registry.subscribe(self._port_provisioned, resources.PORT,
                           provisioning_blocks.PROVISIONING_COMPLETE)
        registry.subscribe(self._handle_segment_change, resources.SEGMENT,
                           events.PRECOMMIT_CREATE)
        registry.subscribe(self._handle_segment_change, resources.SEGMENT,
                           events.PRECOMMIT_DELETE)
        registry.subscribe(self._handle_segment_change, resources.SEGMENT,
                           events.AFTER_CREATE)
        registry.subscribe(self._handle_segment_change, resources.SEGMENT,
                           events.AFTER_DELETE)

        # REVISIT(kent): All the postcommit calls for SG and SG rules are not
        # currently implemented as they are not needed at this moment.
        registry.subscribe(self._handle_security_group_change,
                           resources.SECURITY_GROUP, events.PRECOMMIT_CREATE)
        registry.subscribe(self._handle_security_group_change,
                           resources.SECURITY_GROUP, events.PRECOMMIT_DELETE)
        registry.subscribe(self._handle_security_group_change,
                           resources.SECURITY_GROUP, events.PRECOMMIT_UPDATE)

        # There is no update event to the security_group_rule
        registry.subscribe(self._handle_security_group_rule_change,
                           resources.SECURITY_GROUP_RULE,
                           events.PRECOMMIT_CREATE)
        registry.subscribe(self._handle_security_group_rule_change,
                           resources.SECURITY_GROUP_RULE,
                           events.PRECOMMIT_DELETE)
        try:
            registry.subscribe(self._subnet_delete_precommit_handler,
                    resources.SUBNET, events.PRECOMMIT_DELETE)
            registry.subscribe(self._subnet_delete_after_delete_handler,
                    resources.SUBNET, events.AFTER_DELETE)
        except AttributeError:
            LOG.info(_LI("Detected older version of Neutron, ML2Plus plugin "
                         "is not subscribed to subnet_precommit_delete and "
                         "subnet_after_delete events"))
        self._setup_dhcp()
        self._start_rpc_notifiers()
        self.add_agent_status_check(self.agent_health_check)
        self._verify_service_plugins_requirements()
        self.refresh_network_db_obj = cfg.CONF.ml2plus.refresh_network_db_obj
        self.refresh_port_db_obj = cfg.CONF.ml2plus.refresh_port_db_obj
        self.refresh_subnet_db_obj = cfg.CONF.ml2plus.refresh_subnet_db_obj
        self.refresh_subnetpool_db_obj = (
            cfg.CONF.ml2plus.refresh_subnetpool_db_obj)
        self.refresh_address_scope_db_obj = (
            cfg.CONF.ml2plus.refresh_address_scope_db_obj)
        LOG.info(_LI("Modular L2 Plugin (extended) initialization complete"))

    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
               attributes.SUBNETPOOLS, ['_ml2_md_extend_subnetpool_dict'])

    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
               as_ext.ADDRESS_SCOPES, ['_ml2_md_extend_address_scope_dict'])

    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
               attributes.NETWORKS + '_BULK',
               ['_ml2_md_extend_network_dict_bulk'])

    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
               attributes.SUBNETS + '_BULK',
               ['_ml2_md_extend_subnet_dict_bulk'])

    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
               attributes.SUBNETPOOLS + '_BULK',
               ['_ml2_md_extend_subnetpool_dict_bulk'])

    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
               as_ext.ADDRESS_SCOPES + '_BULK',
               ['_ml2_md_extend_address_scope_dict_bulk'])

    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
               attributes.PORTS + '_BULK',
               ['_ml2_md_extend_port_dict_bulk'])

    def _handle_security_group_change(self, resource, event, trigger,
                                      **kwargs):
        context = kwargs.get('context')
        security_group = kwargs.get('security_group')
        # There is a neutron bug that sometimes it will create a SG with
        # tenant_id field empty. We will not process it further when that
        # happens then.
        if not security_group['tenant_id']:
            return
        original_security_group = kwargs.get('original_security_group')
        mech_context = driver_context.SecurityGroupContext(
            self, context, security_group, original_security_group)
        if event == events.PRECOMMIT_CREATE:
            self._ensure_tenant(context, security_group)
            self.mechanism_manager.create_security_group_precommit(
                mech_context)
            return
        if event == events.PRECOMMIT_DELETE:
            self.mechanism_manager.delete_security_group_precommit(
                mech_context)
            return
        if event == events.PRECOMMIT_UPDATE:
            self.mechanism_manager.update_security_group_precommit(
                mech_context)

    def _handle_security_group_rule_change(self, resource, event, trigger,
                                           **kwargs):
        context = kwargs.get('context')
        if event == events.PRECOMMIT_CREATE:
            sg_rule = kwargs.get('security_group_rule')
            mech_context = driver_context.SecurityGroupRuleContext(
                self, context, sg_rule)
            self.mechanism_manager.create_security_group_rule_precommit(
                mech_context)
            return
        if event == events.PRECOMMIT_DELETE:
            sg_rule = {'id': kwargs.get('security_group_rule_id'),
                       'security_group_id': kwargs.get('security_group_id'),
                       'tenant_id': context.tenant}
            mech_context = driver_context.SecurityGroupRuleContext(
                self, context, sg_rule)
            self.mechanism_manager.delete_security_group_rule_precommit(
                mech_context)

    def _ml2_md_extend_network_dict(self, result, netdb):
        session = patch_neutron.get_current_session()
        with session.begin(subtransactions=True):
            if self.refresh_network_db_obj:
                # In deployment it has been observed that the subnet
                # backref is sometimes stale inside the driver's
                # extend_network_dict. The call to refresh below
                # ensures the backrefs and other attributes are
                # not stale.
                session.refresh(netdb)
            self.extension_manager.extend_network_dict(session, netdb, result)

    def _ml2_md_extend_network_dict_bulk(self, results, _):
        session = patch_neutron.get_current_session()
        with session.begin(subtransactions=True):
            if self.refresh_network_db_obj:
                for _, netdb in results:
                    session.refresh(netdb)
            self.extension_manager.extend_network_dict_bulk(session, results)

    def _ml2_md_extend_port_dict(self, result, portdb):
        session = patch_neutron.get_current_session()
        with session.begin(subtransactions=True):
            if self.refresh_port_db_obj:
                session.refresh(portdb)
            self.extension_manager.extend_port_dict(session, portdb, result)

    def _ml2_md_extend_port_dict_bulk(self, results, _):
        session = patch_neutron.get_current_session()
        with session.begin(subtransactions=True):
            if self.refresh_port_db_obj:
                for _, portdb in results:
                    session.refresh(portdb)
            self.extension_manager.extend_port_dict_bulk(session, results)

    def _ml2_md_extend_subnet_dict(self, result, subnetdb):
        session = patch_neutron.get_current_session()
        with session.begin(subtransactions=True):
            if self.refresh_subnet_db_obj:
                session.refresh(subnetdb)
            self.extension_manager.extend_subnet_dict(
                session, subnetdb, result)

    def _ml2_md_extend_subnet_dict_bulk(self, results, _):
        session = patch_neutron.get_current_session()
        with session.begin(subtransactions=True):
            if self.refresh_subnet_db_obj:
                for _, subnetdb in results:
                    session.refresh(subnetdb)
            self.extension_manager.extend_subnet_dict_bulk(session, results)

    def _ml2_md_extend_subnetpool_dict(self, result, subnetpooldb):
        session = patch_neutron.get_current_session()
        with session.begin(subtransactions=True):
            if self.refresh_subnetpool_db_obj:
                session.refresh(subnetpooldb)
            self.extension_manager.extend_subnetpool_dict(
                session, subnetpooldb, result)

    def _ml2_md_extend_subnetpool_dict_bulk(self, results, _):
        session = patch_neutron.get_current_session()
        with session.begin(subtransactions=True):
            if self.refresh_subnetpool_db_obj:
                for _, subnetpooldb in results:
                    session.refresh(subnetpooldb)
            self.extension_manager.extend_subnetpool_dict_bulk(session,
                                                               results)

    def _ml2_md_extend_address_scope_dict(self, result, address_scopedb):
        session = patch_neutron.get_current_session()
        with session.begin(subtransactions=True):
            if self.refresh_address_scope_db_obj:
                session.refresh(address_scopedb)
            self.extension_manager.extend_address_scope_dict(
                session, address_scopedb, result)

    def _ml2_md_extend_address_scope_dict_bulk(self, results, _):
        session = patch_neutron.get_current_session()
        with session.begin(subtransactions=True):
            if self.refresh_address_scope_db_obj:
                for _, address_scope in results:
                    session.refresh(address_scope)
            self.extension_manager.extend_address_scope_dict_bulk(
                session, results)

    # Base version does not call _apply_dict_extend_functions()
    def _make_address_scope_dict(self, address_scope, fields=None):
        res = {'id': address_scope['id'],
               'name': address_scope['name'],
               'tenant_id': address_scope['tenant_id'],
               'shared': address_scope['shared'],
               'ip_version': address_scope['ip_version']}
        self._apply_dict_extend_functions(as_ext.ADDRESS_SCOPES, res,
                                          address_scope)
        return self._fields(res, fields)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def create_network(self, context, network):
        self._ensure_tenant(context, network[attributes.NETWORK])
        return super(Ml2PlusPlugin, self).create_network(context, network)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def update_network(self, context, id, network):
        return super(Ml2PlusPlugin, self).update_network(context, id, network)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_network(self, context, id):
        return super(Ml2PlusPlugin, self).delete_network(context, id)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def create_network_bulk(self, context, networks):
        self._ensure_tenant_bulk(context, networks[attributes.NETWORKS],
                                 attributes.NETWORK)
        return super(Ml2PlusPlugin, self).create_network_bulk(context,
                                                              networks)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def create_subnet(self, context, subnet):
        self._ensure_tenant(context, subnet[attributes.SUBNET])
        return super(Ml2PlusPlugin, self).create_subnet(context, subnet)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def update_subnet(self, context, id, subnet):
        return super(Ml2PlusPlugin, self).update_subnet(context, id, subnet)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_subnet(self, context, id):
        return super(Ml2PlusPlugin, self).delete_subnet(context, id)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def create_subnet_bulk(self, context, subnets):
        self._ensure_tenant_bulk(context, subnets[attributes.SUBNETS],
                                 attributes.SUBNET)
        return super(Ml2PlusPlugin, self).create_subnet_bulk(context,
                                                             subnets)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def create_port(self, context, port):
        self._ensure_tenant(context, port[attributes.PORT])
        return super(Ml2PlusPlugin, self).create_port(context, port)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def create_port_bulk(self, context, ports):
        self._ensure_tenant_bulk(context, ports[attributes.PORTS],
                                 attributes.PORT)
        return super(Ml2PlusPlugin, self).create_port_bulk(context,
                                                           ports)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def update_port(self, context, id, port):
        return super(Ml2PlusPlugin, self).update_port(context, id, port)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_port(self, context, id, l3_port_check=True):
        return super(Ml2PlusPlugin, self).delete_port(
            context, id, l3_port_check=l3_port_check)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive(context_var_name='plugin_context')
    def get_bound_port_context(self, plugin_context, port_id, host=None,
                               cached_networks=None):
        return super(Ml2PlusPlugin, self).get_bound_port_context(
            plugin_context, port_id, host=host,
            cached_networks=cached_networks)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def update_port_status(self, context, port_id, status, host=None,
                           network=None):
        return super(Ml2PlusPlugin, self).update_port_status(
            context, port_id, status, host=host, network=network)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def port_bound_to_host(self, context, port_id, host):
        return super(Ml2PlusPlugin, self).port_bound_to_host(
            context, port_id, host)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def get_ports_from_devices(self, context, devices):
        return super(Ml2PlusPlugin, self).get_ports_from_devices(
            context, devices)

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def create_subnetpool(self, context, subnetpool):
        self._ensure_tenant(context, subnetpool[attributes.SUBNETPOOL])
        session = context.session
        with session.begin(subtransactions=True):
            result = super(Ml2PlusPlugin, self).create_subnetpool(context,
                                                                  subnetpool)
            self._update_implicit_subnetpool(context, subnetpool, result)
            self.extension_manager.process_create_subnetpool(
                context, subnetpool[attributes.SUBNETPOOL], result)
            mech_context = driver_context.SubnetPoolContext(
                self, context, result)
            self.mechanism_manager.create_subnetpool_precommit(mech_context)
        try:
            self.mechanism_manager.create_subnetpool_postcommit(mech_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("mechanism_manager.create_subnetpool_postcommit "
                              "failed, deleting subnetpool '%s'"),
                          result['id'])
                self.delete_subnetpool(context, result['id'])
        return result

    # REVISIT(rkukura): Is create_subnetpool_bulk() needed?

    @disable_transaction_guard
    @db_api.retry_if_session_inactive()
    def update_subnetpool(self, context, id, subnetpool):
        session = context.session
        with session.begin(subtransactions=True):
            original_subnetpool = super(Ml2PlusPlugin, self).get_subnetpool(
                context, id)
            updated_subnetpool = super(Ml2PlusPlugin, self).update_subnetpool(
                context, id, subnetpool)
            self._update_implicit_subnetpool(context, subnetpool,
                                             updated_subnetpool)
            self.extension_manager.process_update_subnetpool(
                context, subnetpool[attributes.SUBNETPOOL],
                updated_subnetpool)
            mech_context = driver_context.SubnetPoolContext(
                self, context, updated_subnetpool,
                original_subnetpool=original_subnetpool)
            self.mechanism_manager.update_subnetpool_precommit(mech_context)
        self.mechanism_manager.update_subnetpool_postcommit(mech_context)
        return updated_subnetpool

    @disable_transaction_guard
    def delete_subnetpool(self, context, id):
        session = context.session
        with session.begin(subtransactions=True):
            subnetpool = super(Ml2PlusPlugin, self).get_subnetpool(context, id)
            mech_context = driver_context.SubnetPoolContext(
                self, context, subnetpool)
            self.mechanism_manager.delete_subnetpool_precommit(mech_context)
            super(Ml2PlusPlugin, self).delete_subnetpool(context, id)
        self.mechanism_manager.delete_subnetpool_postcommit(mech_context)

    def _update_implicit_subnetpool(self, context, request, result):
        if validators.is_attr_set(request['subnetpool'].get('is_implicit')):
            result['is_implicit'] = request['subnetpool']['is_implicit']
            result['is_implicit'] = (
                self.update_implicit_subnetpool(context, result))

    @disable_transaction_guard
    def create_address_scope(self, context, address_scope):
        self._ensure_tenant(context, address_scope[as_ext.ADDRESS_SCOPE])
        session = context.session
        with session.begin(subtransactions=True):
            result = super(Ml2PlusPlugin, self).create_address_scope(
                context, address_scope)
            self.extension_manager.process_create_address_scope(
                context, address_scope[as_ext.ADDRESS_SCOPE], result)
            mech_context = driver_context.AddressScopeContext(
                self, context, result)
            self.mechanism_manager.create_address_scope_precommit(
                mech_context)
        try:
            self.mechanism_manager.create_address_scope_postcommit(
                mech_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("mechanism_manager.create_address_scope_"
                              "postcommit failed, deleting address_scope"
                              " '%s'"),
                          result['id'])
                self.delete_address_scope(context, result['id'])
        return result

    # REVISIT(rkukura): Is create_address_scope_bulk() needed?

    @disable_transaction_guard
    def update_address_scope(self, context, id, address_scope):
        session = context.session
        with session.begin(subtransactions=True):
            original_address_scope = super(Ml2PlusPlugin,
                                           self).get_address_scope(context, id)
            updated_address_scope = super(Ml2PlusPlugin,
                                          self).update_address_scope(
                                              context, id, address_scope)
            self.extension_manager.process_update_address_scope(
                context, address_scope[as_ext.ADDRESS_SCOPE],
                updated_address_scope)
            mech_context = driver_context.AddressScopeContext(
                self, context, updated_address_scope,
                original_address_scope=original_address_scope)
            self.mechanism_manager.update_address_scope_precommit(mech_context)
        self.mechanism_manager.update_address_scope_postcommit(mech_context)
        return updated_address_scope

    @disable_transaction_guard
    def delete_address_scope(self, context, id):
        session = context.session
        with session.begin(subtransactions=True):
            address_scope = super(Ml2PlusPlugin, self).get_address_scope(
                context, id)
            mech_context = driver_context.AddressScopeContext(
                self, context, address_scope)
            self.mechanism_manager.delete_address_scope_precommit(mech_context)
            super(Ml2PlusPlugin, self).delete_address_scope(context, id)
        self.mechanism_manager.delete_address_scope_postcommit(mech_context)

    def _ensure_tenant(self, context, resource):
        tenant_id = resource['tenant_id']
        self.mechanism_manager.ensure_tenant(context, tenant_id)

    def _ensure_tenant_bulk(self, context, resources, singular):
        tenant_ids = [resource[singular]['tenant_id']
                      for resource in resources]
        for tenant_id in set(tenant_ids):
            self.mechanism_manager.ensure_tenant(context, tenant_id)

    def _get_subnetpool_id(self, context, subnet):
        # Check for regular subnetpool ID first, then Tenant's implicit,
        # then global implicit.
        ip_version = subnet['ip_version']
        return (
            super(Ml2PlusPlugin, self)._get_subnetpool_id(context, subnet) or
            self.get_implicit_subnetpool_id(context,
                                            tenant=subnet['tenant_id'],
                                            ip_version=ip_version) or
            self.get_implicit_subnetpool_id(context, tenant=None,
                                            ip_version=ip_version))

    # REVISIT(ivar): patching bulk gets for extension performance
    def _make_networks_dict(self, networks, context):
        nets = []
        for network in networks:
            res = {'id': network['id'],
                   'name': network['name'],
                   'tenant_id': network['tenant_id'],
                   'admin_state_up': network['admin_state_up'],
                   'mtu': network.get('mtu', n_const.DEFAULT_NETWORK_MTU),
                   'status': network['status'],
                   'subnets': [subnet['id']
                               for subnet in network['subnets']]}
            res['shared'] = self._is_network_shared(context,
                                                    network.rbac_entries)
            nets.append((res, network))

        # Bulk extend first
        self._apply_dict_extend_functions(attributes.NETWORKS + '_BULK', nets,
                                          None)

        result = []
        for res, network in nets:
            res[api_plus.BULK_EXTENDED] = True
            self._apply_dict_extend_functions(attributes.NETWORKS, res,
                                              network)
            res.pop(api_plus.BULK_EXTENDED, None)
            result.append(self._fields(res, []))
        return result

    @db_api.retry_if_session_inactive()
    def get_networks(self, context, filters=None, fields=None,
                     sorts=None, limit=None, marker=None, page_reverse=False):
        session = context.session
        with session.begin(subtransactions=True):
            marker_obj = self._get_marker_obj(context, 'network', limit,
                                              marker)
            query = self._get_collection_query(context, models_v2.Network,
                                               filters=filters,
                                               sorts=sorts,
                                               limit=limit,
                                               marker_obj=marker_obj,
                                               page_reverse=page_reverse)

            net_data = self._make_networks_dict(query, context)
            nets = [attributes.populate_project_info(net) for net in net_data]
            if limit and page_reverse:
                nets.reverse()
            self.type_manager.extend_networks_dict_provider(context, nets)
            for net in nets:
                net[api.MTU] = self._get_network_mtu(net)
            nets = self._filter_nets_provider(context, nets, filters)

        return [self._fields(net, fields) for net in nets]
